#![no_main]
#![no_std]

extern crate cortex_m;
#[macro_use(entry, exception)]
extern crate cortex_m_rt as rt;
extern crate panic_semihosting;

extern crate stm32f4xx_hal as hal;
// #[macro_use(block)]
// extern crate nb;

use core::cell::RefCell;

use cortex_m::interrupt::{free, Mutex};
use cortex_m::peripheral::NVIC;
use hal::gpio::Pull;
use hal::pac::{self, SPI2};
use hal::spi::{Event, Mode, Phase, Polarity, Spi, SpiSlave};
use rf4463::config::RADIO_CONFIG_500_2;
use rf4463::Rf4463;
use ringbuf::{Rb, StaticRb};

use crate::hal::{interrupt, prelude::*};
use crate::rt::ExceptionFrame;

enum SlaveCmd {
    BufferStatus,
    SendPacket,
}

impl SlaveCmd {
    pub fn from_u8(i: u8) -> Option<Self> {
        match i {
            0 => Some(Self::BufferStatus),
            1 => Some(Self::SendPacket),
            _ => None,
        }
    }
}

enum SlaveState {
    Idle,
    BufferStatus,
    RecvPacket(usize),
}

static SPI_SLAVE: Mutex<RefCell<Option<SpiSlave<SPI2, false, u8>>>> =
    Mutex::new(RefCell::new(None));

static SLAVE_STATE: Mutex<RefCell<SlaveState>> = Mutex::new(RefCell::new(SlaveState::Idle));

static BUF: Mutex<RefCell<Option<StaticRb<u8, 16384>>>> = Mutex::new(RefCell::new(None));

const MODE: Mode = Mode {
    polarity: Polarity::IdleLow,
    phase: Phase::CaptureOnFirstTransition,
};

const PACKET_LEN: usize = 256 + 2 + 65;

#[entry]
fn main() -> ! {
    let cp = cortex_m::Peripherals::take().unwrap();
    let dp = pac::Peripherals::take().unwrap();

    let rcc = dp.RCC.constrain();

    let gpioa = dp.GPIOA.split();
    let gpiob = dp.GPIOB.split();

    // setup clocks
    let clocks = rcc.cfgr.use_hse(25.MHz()).sysclk(48.MHz()).freeze();

    // setup 4463 spi
    let mosi = gpioa.pa7.into_alternate();
    let miso = gpioa.pa6.into_alternate();
    let sclk = gpioa.pa5.into_alternate();

    let sdn = gpioa.pa1.into_push_pull_output();

    let cs = gpioa.pa4.into_push_pull_output();

    let spi = Spi::new(dp.SPI1, (sclk, miso, mosi), MODE, 10.MHz(), &clocks);

    let delay = dp.TIM2.delay_ms(&clocks);
    let mut radio = Rf4463::new(spi, sdn, cs, delay, &RADIO_CONFIG_500_2).unwrap();

    // setup Pi SPI (we're the slave)
    let sck2 = gpiob.pb13.internal_resistor(Pull::Up);
    let miso2 = gpiob.pb14.internal_resistor(Pull::Down);
    let mosi2 = gpiob.pb15.internal_resistor(Pull::Down);
    let nss = gpiob.pb12.internal_resistor(Pull::Up).into(); // TODO

    let mut spi2 = dp.SPI2.spi_slave((sck2, miso2, mosi2, Some(nss)), MODE);

    // need something in the outgoing buffer for when the master toggles the clock
    spi2.write(&[0x00]).unwrap();

    spi2.listen(Event::Rxne);

    free(|cs| {
        BUF.borrow(cs).replace(Some(StaticRb::default()));
        SPI_SLAVE.borrow(cs).replace(Some(spi2));
    });

    // enable spi2 interrupt
    unsafe {
        NVIC::unmask(interrupt::SPI2);
    }

    let mut data = [0; PACKET_LEN];
    let mut i = 0;

    loop {
        if i == PACKET_LEN {
            radio.tx(&data).unwrap();
            i = 0;
            continue;
        }

        free(|cs| {
            let mut buf = BUF.borrow(cs).borrow_mut();
            let buf = buf.as_mut().unwrap();

            if let Some(b) = buf.pop() {
                data[i] = b;

                i += 1;

                if buf.free_len() == 16000 {
                    //block!(USART.borrow(cs).borrow_mut().as_mut().unwrap().write(b'E')).unwrap();
                }
            }
        });
    }
}

#[exception]
unsafe fn HardFault(ef: &ExceptionFrame) -> ! {
    panic!("{:#?}", ef);
}

#[interrupt]
fn SPI2() {
    free(|cs| {
        let mut state = SLAVE_STATE.borrow(cs).borrow_mut();

        let mut spi = SPI_SLAVE.borrow(cs).borrow_mut();
        let spi = spi.as_mut().unwrap();

        let mut buf = BUF.borrow(cs).borrow_mut();
        let buf = buf.as_mut().unwrap();

        let mut x = [0];
        spi.read(&mut x).unwrap();

        let out = match *state {
            SlaveState::Idle => {
                if let Some(cmd) = SlaveCmd::from_u8(x[0]) {
                    match cmd {
                        SlaveCmd::BufferStatus => {
                            *state = SlaveState::BufferStatus;

                            if buf.free_len() < PACKET_LEN {
                                0x01 // not enough space for a new packet
                            } else {
                                0x02 // safe to send a packet
                            }
                        }

                        SlaveCmd::SendPacket => {
                            *state = SlaveState::RecvPacket(PACKET_LEN - 1);

                            0x00
                        }
                    }
                } else {
                    0x00
                }
            }
            SlaveState::BufferStatus => {
                *state = SlaveState::Idle;

                0x00
            }
            SlaveState::RecvPacket(i) => {
                *state = if i == 0 {
                    SlaveState::Idle
                } else {
                    SlaveState::RecvPacket(i - 1)
                };

                buf.push(x[0]).unwrap();

                0x00
            }
        };

        spi.write(&[out]).unwrap();
    });
}
